function [M,E0,Eiter] = DMRG_3S_Ex(Hs,Nkeep,Nsweep,varargin)
% < Description >
%
% [M,E0,Eiter] = DMRG_3S_Ex(Hs,Nkeep,Nsweep [, option])
%
% Strictly single-site density-matrix renormalization group (DMRG3S) 
% calculation to search for the ground state and its energy of a
% one-dimensional system, whose Hamiltonian is given by the matrix product 
% operator Hs.
% For reference, see C. Hubig, PRB 91, 155115 (2015)
%
% < Input >
% Hs : [1 x N cell array] Matrix product operator (MPO) of the Hamiltonian.
%       Each Hs{n} acts on site n, and is rank-4 tensor. The order of legs
%       of Hs{n} is left-bottom-right-top, where bottom (top) leg is to be
%       contracted with bra (ket) tensor. The length N of Hs, i.e.,
%       numel(Hs), determines the chain length.
% Nkeep : [numeric] Maximum bond dimension of the matrix product state
%       (MPS) to consider.
% Nsweep : [numeric] Number of sweeps will be 2 * Nsweep, as there are
%       Nsweep times of round trip (right -> left, left -> right).
%
% < Option >
% 'rand' : If given, the MPS is initialized by random tensors. If not
%       (which is default), it is initialized by the result of the
%       iterative diagonalization from the left.
% 'Minit', ..: [1 x N cell array] initial MPS
% 'noex' : If given, no subspace expansion will be performed, i.e.
%           conventional single-site DMRG is run
%
% < Output >
% M : [1 x N cell array] The result MPS which is obtained variationally to
%       have the minimum expectation value of Hamiltonian H. It is in 
%       *left-canonical* form, since the last sweep is from left to right.
% E0 : [numeric] The energy of M.
% Eiter : [N x (2*Nsweep) numeric array] Each element Eiter(m,n) means the
%       variational energy in the m-th iteration within the n-th sweep.
%       Odd n is for right-to-left sweep and even n for left-to-right
%       sweep. Note that the iteration index m matches with the site index
%       for left-to-right sweep; the iteration m corresponds to the site
%       (N+1-m) for right-to-left sweep.
%
% Written by A.Gleis (May 24,2021), based on DMRG_1site by S.Lee


tobj = tic2;

% default value of optional input
Minit = [];
isrand = false; % true: initialize M with random numbers
isex = true;

% parsing option
while ~isempty(varargin)
    switch varargin{1}
        case 'rand'
            isrand = true;
            varargin(1) = [];
        case 'Minit'
            Minit = varargin{2};
            varargin(1:2) = [];
        case 'noex'
            isex = false;
            varargin(1) = [];
        otherwise
            error('ERR: check input!');
    end
end
    
% % sanity check for input and option
N = numel(Hs);

if N < 2
    error('ERR: chain is too short.');
elseif ~isempty(Minit) && (~iscell(Minit) || (numel(Minit) ~= N))
    error('ERR: Initial MPS Minit has different lengths from that of Hs.');
end

for itN = (1:N)
    if size(Hs{itN},2) ~= size(Hs{itN},4)
        error(['ERR: The second and fourth legs of Hs{', ...
            sprintf('%i',itN),'} have different dimensions.']);
    end
end
% % %

% show message
fprintf('DMRG3S: search for the ground state\n');
fprintf(['# of sites = ',sprintf('%i',numel(Hs)), ...
    ', Nkeep = ',sprintf('%i',Nkeep),', # of sweeps = ',sprintf('%i',Nsweep),' x 2\n']);


M = cell(1,N);
if ~isempty(Minit)
    M = Minit;
elseif isrand
    % initialize with random MPS
    for itN = (1:N)
        if itN == 1
            M{itN} = rand(1,size(Hs{itN},4),Nkeep); % left leg is dummy
        elseif itN == N
            M{itN} = rand(Nkeep,size(Hs{itN},4),1); % right leg is dummy
        else
            M{itN} = rand(Nkeep,size(Hs{itN},4),Nkeep);
        end
    end
else
    % Initilize with the result of the iterative diagonalization from the
    % left.
    % Here the input Hamiltonian is defined by MPO, so the Hamiltonian for
    % a part of the chain has an open leg to the right. To diagonalize the
    % Hamiltonian, we choose the index 1 for the right leg. It is
    % consistent with the convention by which we define the Hamiltonian for
    % the last chain site.
     
    Hzip = 1; % MPO Hamiltonian contracted by MPS kets/bras, "zipped" from 
              % the left end; initialize as 1 for dummy leg
    for itN = (1:N)
        if itN == 1
            M{itN} = getIdentity(1,3,Hs{itN},4); % left leg is dummy
        else
            M{itN} = getIdentity(M{itN-1},3,Hs{itN},4);
        end
        Hzip = updateLeft(Hzip,3,M{itN},Hs{itN},4,M{itN});
        % Hamiltonian to be used for the iterative diagonalization
        Hzip2 = reshape(Hzip(:,1,:),[size(Hzip,1) size(Hzip,3)]); 
        % here we chose the index 1 for the 2nd leg, following the
        % convention of defining the right-most tensor within the MPO
        % Hamiltonian.
        [V,D] = eig((Hzip2+Hzip2')/2);
        [~,ids] = sort(diag(D),'ascend');
        V = V(:,ids);
        if itN < N
            V = V(:,(1:min(size(V,2),Nkeep)));
        else
            V = V(:,1);
        end
        M{itN} = contract(M{itN},3,3,V,2,1);
        Hzip = contract(Hzip,3,3,V,2,1);
        Hzip = contract(conj(V),2,1,Hzip,3,1);
    end
end

M = canonForm(M,0); % bring into right-canonical form
M = canonForm(M,N); % bring into left-canonical form

% ground-state energy for each iteration
Eiter = zeros(N,2*Nsweep);
% later, Eiter(end,end) will be taken as the final result E0

% % Hamiltonian for the left/right parts of the chain
Hlr = cell(1,N+2);
% Hlr{1} and Hlr{end} are dummies; they will be kept empty. These dummies
% are introduced for convenience.

% Since M is in left-canonical form by now, Hlr{..} are the left parts of
% the Hamiltonian. That is, Hlr{n+1} is the left part of Hamiltonian which
% is obtained by contracting M(1:n) with Hs(1:n). (Note the index for Hlr
% is n+1, not n, since Hlr{1} is dummy.)
for itN = (1:N)
    if itN == 1
        % "remove" the left leg (1st leg) which is dummy by permuting to the last
        H2 = permute(Hs{itN},[2 3 4 1]);
        Hlr{itN+1} = updateLeft([],[],M{itN},H2,3,M{itN});
    else
        Hlr{itN+1} = updateLeft(Hlr{itN},3,M{itN},Hs{itN},4,M{itN});
    end
end

if ~isempty(Minit)
    disptime(['Initial MPS given as input. Energy = ',sprintf('%.7g',Hlr{N+1})]);
elseif isrand
    disptime(['Initialize with random MPS. Energy = ',sprintf('%.7g',Hlr{N+1})]);
else
    disptime(['Initialize with iterative diagonalization. Energy = ',sprintf('%.7g',Hlr{N+1})]);
end

for itS = (1:Nsweep)
    % right -> left
    for itN = (N:-1:1)
        % Use eigs_1site to obtain the variationally chosen ket tensor
        % Aeff and energy expectation value Eeff
        [Aeff,Eeff] = eigs_1site(Hlr{itN},Hs{itN},Hlr{itN+2},M{itN});
        
        Eiter(N+1-itN,2*itS-1) = Eeff;

        % subspace expansion step:
        if itN > 1 && isex...
            && size(Aeff,1) < size(Aeff,2)*size(Aeff,3) && ...
            size(M{itN-1},3) < size(M{itN-1},2)*size(M{itN-1},1) 
            
            % procede as in 2-site DMRG but only do one Lanczos step:
            
            Hloc = contract(Hs{itN-1},4,3,Hs{itN},4,1,[1 2 4 5 3 6]);
            % merge the physical legs of Hloc by reshaping
            Hloc = reshape(Hloc,[size(Hloc,1),size(Hloc,2)*size(Hloc,3), ...
                size(Hloc,4),size(Hloc,5)*size(Hloc,6)]);

            % merge two rank-3 ket tensors to a single rank-3 tensor
            Ain = contract(M{itN-1},3,3,Aeff,3,1);
            % merge the physical legs of Ain by reshaping
            Ain = reshape(Ain,[size(Ain,1),size(Ain,2)*size(Ain,3),size(Ain,4)]);

            % Use eigs_1site for the Lanczos step
            % Aeff and energy expectation value Eeff
            [Aeff,Eeff] = eigs_1site(Hlr{itN-1},Hloc,Hlr{itN+2},Ain,'N',1);
            
            Aeff = reshape(Aeff,[size(Aeff,1),size(Hs{itN-1},4),size(Hs{itN},4),size(Aeff,3)]);
            % update M{itN-1} and M{itN} by using Aeff, via SVD
            % decompose Aeff
            [UT,ST,M{itN}] = svdTr(Aeff,4,[1 2],Nkeep,[]);
            ST = ST./sqrt(sum(ST.^2));
            M{itN-1} = contract(UT,3,3,diag(ST),2,1);
            
            % update energy
            Eiter(N+1-itN,2*itS-1) = Eeff;
            % end of subspace expansion
        else
            % update M{itN} and M{itN-1} by using Aeff, via SVD
            % decompose Aeff
            [UT,ST,M{itN}] = svdTr(Aeff,3,1,Nkeep,[]);
            % contract UT*ST with M{itN}, to update M{itN}
            if itN > 1
                M{itN-1} = contract(M{itN-1},3,3,UT*diag(ST),2,1);
            end
        end
        
        % update the Hamiltonian in effective basis
        T = permute(M{itN},[3 2 1]); % permute left<->right, to make use of updateLeft
        if itN == N
            % "remove" the right leg (3rd leg) of Hs{itN}, which is dummy,
            % by permuting to the last; MATLAB automatically suppresses the
            % trailing singleton dimensions
            H2 = permute(Hs{itN},[2 1 4 3]); % bottom-left-top (-right)
            Hlr{itN+1} = updateLeft([],[],T,H2,3,T);
        elseif itN == 1
            Hlr{itN+1} = [];
        else
            % permute left<->right for Hs{itN} as well, to make use of updateLeft
            H2 = permute(Hs{itN},[3 2 1 4]); % right-bottom-left-top
            Hlr{itN+1} = updateLeft(Hlr{itN+2},3,T,H2,4,T);
        end
        
    end
    % display informaiton of the sweep
    str = ['Sweep #',sprintf('%i/%i',[2*itS-1,2*Nsweep]),' (right -> left) : Energy = ',sprintf('%.7g',Eiter(N,2*itS-1))];
    disptime(str);

    % left -> right
    for itN = (1:N)
        [Aeff,Eeff] = eigs_1site(Hlr{itN},Hs{itN},Hlr{itN+2},M{itN});
        
        Eiter(itN,2*itS) = Eeff;
        
        % subspace expansion step:
        if itN < N && isex && ...
            size(Aeff,3) < size(Aeff,2)*size(Aeff,1) && ...
                size(M{itN+1},1) < size(M{itN+1},2)*size(M{itN+1},3)
            
            % % % TODO - Exercise 1: % % %
            % implement the subspace expansion + 1-step Lanczos as above
            % procede as for 2-site DMRG but with only one Lanczos step

            
            
            % % % TODO - Exercise 1 - end % % %

        else
            % update M{itN} and M{itN+1} by using Aeff, via SVD
            % decompose Aeff
            [M{itN},ST,VT] = svdTr(Aeff,3,[1 2],Nkeep,[]);
            % contract UT*ST with M{itN}, to update M{itN}
            if itN < N
                M{itN+1} = contract(diag(ST)*VT,2,2,M{itN+1},3,1);
            end
        end
        
        % update the Hamiltonian in effective basis
        if itN == 1
            % "remove" the left leg (1st leg) of Hs{itN}, which is dummy,
            % by permuting to the last; MATLAB automatically suppresses the
            % trailing singleton dimensions
            H2 = permute(Hs{itN},[2 3 4 1]); % bottom-right-top (-left)
            Hlr{itN+1} = updateLeft([],[],M{itN},H2,3,M{itN});
        elseif itN == N
            Hlr{itN+1} = [];
        else
            Hlr{itN+1} = updateLeft(Hlr{itN},3,M{itN},Hs{itN},4,M{itN});
        end
        
    end
    
    % display informaiton of the sweep
    str = ['Sweep #',sprintf('%i/%i',[2*itS,2*Nsweep]),' (left -> right) : Energy = ',sprintf('%.7g',Eiter(N,2*itS))];
    disptime(str);
end

E0 = Eiter(N,2*itS); % take the last value
    
toc2(tobj,'-v');

end