%% [Tutorial T09.2] DMRG: subspace expansion and error estimation
% Author: Andreas Gleis
%% Exercise (a): Complete the subspace expansion feature of D|MRG_3S_Ex.m|
% There is a function |DMRG_3S_Ex.m| which is placed in the same subdirectory 
% together with this tutorial material. This function partially implements subspace 
% expansion in the single-site update for searching the ground state. The subspace 
% expansion prevents the single-site DMRG algoritm to get stuck in local minima 
% and in principle allows to dynamically adjust the bond-dimension, similar to 
% the 2-site DMRG. Nevertheless, it remains a strictly single-site (3S) update, 
% which has performance advantages in case the local dinemsion $d$ becomes large. 
% *Complete the parts which are enclosed by the comments |TODO - Exercise 1.*|
%% Majumdar-Ghosh model
% We have already encountered the Majumdar-Ghosh model in T07*.* The Majumdar-Ghosh 
% model is a spin-1/2 chain with nearest-neighbor and next-nearest-neighbor interactions,
% 
% $$H = \sum_{\ell=1}^{N-1} \hat{\vec{S}}_\ell \cdot \hat{\vec{S}}_{\ell+1} 
% + \frac{1}{2} \sum_{\ell=1}^{N-2} \hat{\vec{S}}_\ell \cdot \hat{\vec{S}}_{\ell+2} 
% .$$
% 
% It has the exact ground state, with energy $E_0 = -3N/8$ and the bond dimension 
% $N_\mathrm{keep} = 2$. In the following, we will compare the ground-state search 
% of the single-site DMRG with and without subspace expansion using $N_\mathrm{keep} 
% = 10$ and a series of special initial states for which the single-site DMRG 
% algorithm is prone to getting stuck. Note that the option 'noex' switches off 
% the subspace expansion.

clear

% system parameter
J1 = 1; % nearest-neighbour coupling strength
J2 = 1/2; % next-nearest-neighbour coupling strength
N = 40; % number of sites in a chain

% DMRG parameter
Nkeep = 10; % bond dimension
Nsweep = 5; % number of pairs of left+right sweeps

% Local operators
[S,I] = getLocalSpace('Spin',1/2);

% Hamiltonian tensor for each chain site
Hloc = cell(8,8);
Hloc(:) = {zeros(size(I))};
Hloc{1,1} = I;
Hloc{2,1} = squeeze(S(:,1,:));
Hloc{3,1} = squeeze(S(:,2,:));
Hloc{4,1} = squeeze(S(:,3,:));
Hloc{5,2} = I;
Hloc{6,3} = I;
Hloc{7,4} = I;
Hloc{8,2} = J1*(Hloc{2,1}');
Hloc{8,3} = J1*(Hloc{3,1}');
Hloc{8,4} = J1*(Hloc{4,1}');
Hloc{8,5} = J2*(Hloc{2,1}');
Hloc{8,6} = J2*(Hloc{3,1}');
Hloc{8,7} = J2*(Hloc{4,1}');
Hloc{end,end} = I;
Hloc = cell2mat(reshape(Hloc,[1 1 size(Hloc,1) size(Hloc,2)]));
Hloc = permute(Hloc,[3 1 4 2]); % leg order: left-bottom-right-top

% full chain
Hs = cell(1,N);
Hs(:) = {Hloc};
Hs{1} = Hs{1}(end,:,:,:); % choose the last components of the left leg
Hs{end} = Hs{end}(:,:,1,:); % choose the first components of the right leg

E0_exact = -3*N/8; % exact value
%% 
% After setting up the Hamiltonian an choosing our parameters, we choose an 
% initial state with $A^{\uparrow}_{11} = A^{\downarrow}_{22} = 1$ while all other 
% entries are set to zero. At the left boundary, we choose all entries zero except 
% $A^{\uparrow}_{11} = A^{\downarrow}_{12} = 1$ while on the right boundary, only  
% $A^{\uparrow}_{11} = A^{\downarrow}_{21} = 1$ are non-zero. (*Quick question:* 
% what kind of state does this represent?) We then perturb this state by adding 
% small random numbers to the $A$-matrices. We iterate over several randomly generated 
% perturbations and keep track how single-site DMRG converges compared to DMRG3S.

del = 1e-8; % small number
Nr = 10; % number of DMRG runs with random perturbation
MA = cell(1,N);
A = zeros(Nkeep,2,Nkeep);
A(1,1,1) = 1;
A(2,2,2) = 1;
MA(:) = {A};
MA{1} = A(1,:,:)+A(2,:,:);
MA{end} = A(:,:,1)+A(:,:,2);

% worst difference between GS energy and mean energy of the optimized MPS:
dEGS = 0;
MR = cell(1,N); % corresponding random perturbation
Eiter_1site = cell(1,Nr);
Eiter_3S = cell(1,Nr);

for itr = 1:Nr
    Minit = cell(1,N);
    for itN = 1:N
        if itN == 1
            Minit{itN} = rand(1,2,Nkeep)*del+MA{itN};
        elseif itN == N
            Minit{itN} = rand(Nkeep,2,1)*del+MA{itN};
        else
            Minit{itN} = rand(Nkeep,2,Nkeep)*del+MA{itN};
        end
    end
    
    [~,~,Eiter_1site{itr}] = DMRG_3S_Ex(Hs,Nkeep,Nsweep,'Minit',Minit,'noex');
    [~,~,Eiter_3S{itr}] = DMRG_3S_Ex(Hs,Nkeep,Nsweep,'Minit',Minit);
    
end
% visualize:
figure;
hold on;
Cls = parula(Nr);
for itr = 1:Nr
    plot((1:numel(Eiter_3S{itr}))/N,abs(Eiter_3S{itr}(:)-E0_exact),...
        'LineWidth',1,'Color',Cls(itr,:));
    plot((1:numel(Eiter_1site{itr}))/N,abs(Eiter_1site{itr}(:)-E0_exact),...
        '--','LineWidth',1,'Color',Cls(itr,:));
end
legend({'DMRG3S','single-site DMRG'},'FontSize',13);
set(gca,'FontSize',13,'LineWidth',1,'YScale','log');
xlim([0 2*Nsweep]);
grid on;
xlabel('# of sweeps');
ylabel('Ground-state energy error');
hold off;
%% 
% We see that DMRG3S converges up to numerical precision after around 6 sweeps 
% while conventional single-site DMRG takes more sweeps to converge. For some 
% of the random perturbations, single-site DMRG does not converge to the ground-state 
% at all, opposed to DMRG3S.
%% Exercise (b): Complete the function twoSiteVariance|_Ex.m|
% There is a function twoSiteVariance|_Ex.m| which is placed in the same subdirectory 
% together with this tutorial material. This function partially implements the 
% calculation of the two-site variance as an error meassure. *Complete the parts 
% which are enclosed by the comments |TODO - Exercise 2.*|
%% 
% 
%% Heisenberg chain: error estimation
% To test twoSiteVariance|_Ex.m| we consider a Heisenberg spin chain and calculate 
% its ground-state. Its Hamiltonian is given by
% 
% $$H = \sum_{l=1}^{N-1} \vec{S}_{l}\vec{S}_{l+1} \, .$$
% 
% In the following, we will compare the two-site variance of the optimized GS 
% obtained by DMRG3S and two-site DMRG for different bond dimensions.

clear

% system parameter
J = 1; % coupling strength
N = 40; % number of sites in a chain

% DMRG parameter
Nsweep = 10; % number of pairs of left+right sweeps

% magnitude of local spin
s = 1/2;
% Local operators
[S,I] = getLocalSpace('Spin',s);
%% 
% We set up the MPO of the Heisenberg chain as given in the lecture:

% % MPO formulation of Hamiltonian
% Hamiltonian tensor for each chain site
Hloc = cell(5,5);
Hloc(:) = {zeros(size(I))};
Hloc{1,1} = I;
Hloc{2,1} = squeeze(S(:,1,:));
Hloc{3,1} = squeeze(S(:,2,:));
Hloc{4,1} = squeeze(S(:,3,:));
Hloc{end,2} = J*(Hloc{2,1}');
Hloc{end,3} = J*(Hloc{3,1}');
Hloc{end,4} = J*(Hloc{4,1}');
Hloc{end,end} = I;
Hloc = cell2mat(reshape(Hloc,[1 1 size(Hloc,1) size(Hloc,2)]));
Hloc = permute(Hloc,[3 1 4 2]); % leg order: left-bottom-right-top

% full chain
Hs = cell(1,N);
Hs(:) = {Hloc};
Hs{1} = Hs{1}(end,:,:,:); % choose the last components of the left leg
Hs{end} = Hs{end}(:,:,1,:); % choose the first components of the right leg
%% 
% Run the DMRG for different bond dimensions.

Nkeeps = [5,10,20,40,80,160]; % bond dimensions
var_2site = zeros(1,numel(Nkeeps));
var_3S = zeros(1,numel(Nkeeps));
for itK = 1:numel(Nkeeps)
    Nkeep = Nkeeps(itK);
    
    % first 2-site DMRG:
    [MGS_2site,~,~] = DMRG_2site(Hs,Nkeep,Nsweep);
    var_2site(itK) = twoSiteVariance_Ex(MGS_2site,Hs);
    
    % DMRG3S
    [MGS_3S,~,~] = DMRG_3S_Ex(Hs,Nkeep,Nsweep);
    var_3S(itK) = twoSiteVariance_Ex(MGS_3S,Hs);
end
%% 
% Plot how the two-site variance decreases with the bond dimension:

    figure;
    semilogy(Nkeeps,var_2site,'-x','LineWidth',1);
    hold on;
    semilogy(Nkeeps,var_3S,'--o','LineWidth',1);
    set(gca,'FontSize',13,'LineWidth',1);
    xlim([0 max(Nkeeps)]);
    grid on;
    legend({'2-site DMRG','DMRG3S'},'FontSize',13)
    xlabel('N_{keep}');
    ylabel('two-site variance');
    hold off;
%% 
% As expected, the variance decreases with $N_{\mathrm{keep}}$. Two-site DMRG 
% and DMRG3S yield almost identical variances. Note however that DMRG3S is only 
% marginally faster for $s=1/2$ than 2-site DMRG. This is because of the small 
% physical dimension $d=2$. If $s$ is increased, $d$ increases and the DMRG3S 
% eventually becomes faster. However, calculations also become quite a bit heavier. 
% But if your computer is able to handle it, give larger $s$ a try!